import pickle
from luko_rada_package_1.rada_factories import (
    UkraineRadaFactory,
    PolandRadaFactory
)

class DataBase:
    @classmethod
    def save_to_db(cls, vr):
        confirmance = input('Are you sure? press yes/no: ')
        if confirmance == 'yes':
            with open(cls.db_file, 'wb') as f:
                pickle.dump(vr, f, pickle.HIGHEST_PROTOCOL)
                print('DATA has been recorded!')

    @classmethod
    def load_from_db(cls):
        with open(cls.db_file, 'rb') as f:
            VR = pickle.load(f)
        return VR


class UkraineDataBase(DataBase):
    db_file = 'db_data/ukraine.pickle'


class PolandDataBase(DataBase):
    db_file = 'db_data/poland.pickle'



from rada_metaclass import MetaDescriptors
from utils import Sorting
from luko_rada_package_1.rada_factories import (
    UkraineRadaFactory,
    PolandRadaFactory
)

class Human(metaclass=MetaDescriptors):
    int_types = ('weight', 'height')

    def __init__(self, weight, height):
        self.weight = weight
        self.height = height


class Deputat(Human):
    str_types = ('last_name', 'first_name')
    int_types = ('age', 'bribe_amount')

    def __init__(self, weight, height, last_name, first_name, age):
        super(Deputat, self).__init__(weight, height)
        self.last_name = last_name
        self.first_name = first_name
        self.age = age

    def __eq__(self, other):
        return self.weight == other.weight and \
               self.height == other.height and \
               self.age == other.age

    def __hash__(self):
        return hash(self.weight) + hash(self.height) + hash(self.age)

    def __str__(self):
        return f'{self.first_name} {self.last_name} {self.weight}kg {self.height}cm {self.age}age'


class UkraineDeputat(Deputat):
    def __init__(self, weight, height, last_name, first_name, age, bribe_taker):
        super(UkraineDeputat, self).__init__(weight, height, last_name, first_name, age)
        self.bribe_taker = bool(int(bribe_taker))
        self.bribe_amount = 0

    def give_tribute(self):
        if self.bribe_taker is False:
            print('this deputy does not take bribes...')
            return False

        money = input('how much grivnjas would you like to give:')
        if int(money) > 10000:
            print('Police will arrest the deputy!')
            return False

        self.bribe_amount += int(money)
        return self.bribe_amount

    @classmethod
    def describe_deputy(cls):
        personal_data = input(
            'Please enter: weight, height, last_name, first_name, age, is_briber?[0 or 1]: '
        )

        pd = personal_data.split(', ')

        try:
            deputy = cls(pd[0], pd[1], pd[2], pd[3], pd[4], pd[5])
            return deputy
        except IndexError as err:
            print('data should be separated by comma and whitespace', 'Invalid Data')


class PolandDeputat(Deputat):
    def __init__(self, weight, height, last_name, first_name, age):
        super(PolandDeputat, self).__init__(weight, height, last_name, first_name, age)

    @classmethod
    def describe_deputy(cls):
        personal_data = input(
            'Please enter: weight, height, last_name, first_name, age: '
        )

        pd = personal_data.split(', ')

        try:
            deputy = cls(pd[0], pd[1], pd[2], pd[3], pd[4])
            return deputy
        except IndexError as err:
            print('data should be separated by comma and whitespace', 'Invalid Data')


class Fraction(metaclass=MetaDescriptors):
    str_types = ('name',)

    def __init__(self, name, deputat_class):
        self.name = name
        self.deputat_class = deputat_class
        self.deputies = set()

    def __str__(self):
        return f'{self.name} members: {len(self.deputies)}'

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)

    def add_deputy(self):
        deputy = self.deputat_class.describe_deputy()

        if not deputy:
            print('please describe deputy correctly, Invalid Data')
            return False

        if deputy not in self.deputies:
            self.deputies.add(deputy)
            print(f'{deputy} has been added to the fraction; {self.name}')
            return deputy
        else:
            print(f'{deputy} is already a participant of the fraction {self.name}')

    def remove_deputy(self, deputy=None):
        if not deputy:
            personal_data = input(
                'Please enter: weight, height, last_name, first_name, age, is_briber?[True or False]: '
            )
            pd = personal_data.split(', ')

            # should work fine if Deputat class is hashed. expected equal() to return True.
            try:
                deputy = Deputat(pd[0], pd[1], pd[2], pd[3], pd[4], bool(pd[5]))
            except IndexError as err:
                print('data should be separated by comma and whitespace', 'Invalid Data')
                return False

        if deputy in self.deputies:
            self.deputies.remove(deputy)
            print(f'{deputy} has been excluded from the fraction: {self.name}')
        else:
            print(f'{deputy} is not a participant of the fraction: {self.name}')

    def find_all_deputies(self):
        deputies = list(self.deputies)
        deputies.sort(key=Sorting('first_name', 'last_name'))

        print('List of all deputies...')
        for deputy in deputies:
            print(deputy)

    def clean_out_deputies(self):
        print('All deputies has been cleaned out')
        self.deputies = set()

    def is_participant(self, vr, deputy):
        for fraction in vr.fractions:
            if deputy in fraction.deputies:
                print(f'{deputy} is a member of the {self.name} fraction')
                return True

        print(f'{deputy} is not a member of the {self.name} fraction')
        return False


class UkraineFraction(Fraction):

    def __init__(self, name, deputat_class):
        super(UkraineFraction, self).__init__(name, deputat_class)

    def find_all_bribe_takers(self):
        print('The following bribe_takers have been found...')
        deputies = [dep for dep in self.deputies if dep.bribe_taker]
        deputies.sort(key=Sorting('bribe_amount'))

        for deputy in deputies:
            print(deputy, 'grivnjas taken: {}'.format(deputy.bribe_amount))

    def find_highest_bribe_taker(self):
        print('The most greedy deputy is:')
        deputies = [dep for dep in self.deputies if dep.bribe_taker]
        deputies.sort(key=Sorting('bribe_amount'))

        print(deputies[-1], 'grivnjas taken: {}'.format(deputies[-1].bribe_amount))


class PolandFraction(Fraction):

    def __init__(self, name, deputat_class):
        super(PolandFraction, self).__init__(name, deputat_class)


class VerkhovnaRada:

    def __init__(self):
        self.fractions = set()

    def add_fraction(self, fraction):
        if fraction not in self.fractions and fraction.name:
            print(f'Fraction {fraction.name} has been added to VR')
            self.fractions.add(fraction)
            return True

        print('already in VR')

    def remove_fraction(self, fraction):
        if fraction in self.fractions:
            print(f'Fraction {fraction.name} has been removed from VR')
            self.fractions.remove(fraction)
            return True

        print(f'Fraction {fraction.name} not in the list.')
        return False

    def print_all_fractions(self):
        print('List of all fractions:')
        for fr in self.fractions:
            print(fr)

    @staticmethod
    def print_fraction(fraction):
        print(fraction)

    @staticmethod
    def add_deputy_to_fraction(deputy, fraction):
        print(f'Adding deputy: {deputy} ...')
        fraction.add_deputy(deputy)

    def remove_deputy_from_fraction(self, fraction, deputy):
        print(f'Removing deputy: {deputy} ...')
        for fr in self.fractions:
            if fraction == fr:
                fr.remove_deputy(deputy)
                return

    def print_all_deputies(self):
        for fraction in self.fractions:
            print('{}: '.format(fraction.name))
            for deputy in fraction.deputies:
                print(''.rjust(len(fraction.name)), deputy)

    def is_present_in_vr(self, deputy):
        for fraction in self.fractions:
            if deputy in fraction:
                print('True')
        print('False')

    def update_deputy(self, deputy):
        for fraction in self.fractions:
            for dp in fraction.deputies:
                if deputy == dp:
                    dp.bribe_amount = deputy.bribe_amount


class UkraineVerkhovnaRada(VerkhovnaRada):
    name = 'Verkhovna Rada Ukraine'

    def __init__(self):
        super(UkraineVerkhovnaRada, self).__init__()

    def print_all_bribe_takers(self):
        print('Printing all bribe takers in VR')
        for fraction in self.fractions:
            for deputy in fraction.deputies:
                if deputy.bribe_taker:
                    print(deputy, 'grivnjas taken: {}'.format(deputy.bribe_amount))

    def higher_bribe_taker(self):
        bribe_takers = []
        for fraction in self.fractions:
            for deputy in fraction.deputies:
                if deputy.bribe_taker:
                    bribe_takers.append(deputy)

        bribe_takers.sort(key=Sorting('bribe_amount'))
        print(bribe_takers[-1], 'grivnjas taken: {}'.format(bribe_takers[-1].bribe_amount))

    def update_deputy(self, deputy):
        for fraction in self.fractions:
            for dp in fraction.deputies:
                if deputy == dp:
                    dp.bribe_amount = deputy.bribe_amount


class PolandVerkhovnaRada(VerkhovnaRada):
    name = 'Verkhovna Rada Poland'

    def __init__(self):
        super(PolandVerkhovnaRada, self).__init__()




from abc import ABC, abstractmethod

from rada_databases import UkraineDataBase, PolandDataBase
from rada_models import (
    UkraineVerkhovnaRada,
    PolandVerkhovnaRada,
    UkraineFraction,
    PolandFraction,
    UkraineDeputat,
    PolandDeputat,
)
from rada_consoles import UkraineConsole, PolandConsole


class RadaFactory(ABC):

    @abstractmethod
    def __init__(self):
        pass


class UkraineRadaFactory(RadaFactory):

    def __init__(self):
        self.rada_class = UkraineVerkhovnaRada
        self.fraction_class = UkraineFraction
        self.deputat_class = UkraineDeputat
        self.console_class = UkraineConsole
        self.db_class = UkraineDataBase

    def __call__(self, *args, **kwargs):
        rada = self.rada_class()
        return UkraineConsole(rada, self.fraction_class, self.deputat_class, self.db_class)


class PolandRadaFactory(RadaFactory):

    def __init__(self):
        self.rada_class = PolandVerkhovnaRada
        self.fraction_class = PolandFraction
        self.deputat_class = PolandDeputat
        self.console_class = PolandConsole
        self.db_class = PolandDataBase

    def __call__(self, *args, **kwargs):
        rada = self.rada_class()
        return PolandConsole(rada, self.fraction_class, self.deputat_class, self.db_class)


class IntDescriptor:
    def __set_name__(self, owner, name):
        self.name = '_' + name

    def __get__(self, instance, owner):
        return getattr(instance, self.name)

    def __set__(self, instance, value):
        value = int(value)
        assert isinstance(value, int) \
               and value > -1
        setattr(instance, self.name, value)

    def __delete__(self, instance):
        delattr(instance, self.name)


class StringDescriptor:
    def __set_name__(self, owner, name):
        self.name = '_' + name

    def __get__(self, instance, owner):
        return getattr(instance, self.name)

    def __set__(self, instance, value):
        assert isinstance(value, str) \
               and len(value.strip()) > 2
        setattr(instance, self.name, value)

    def __delete__(self, instance):
        delattr(instance, self.name)


from abc import ABC, abstractmethod
from luko_rada_package_1.rada_factories import (
    UkraineRadaFactory,
    PolandRadaFactory
)

class Console(ABC):
    def __init__(self, rada, fraction_class, deputat_class, db_class):
        self.rada = rada
        self.fraction_class = fraction_class
        self.deputat_class = deputat_class
        self.db = db_class
    
    @abstractmethod
    def run(self):
        pass
    
    
class UkraineConsole(Console):
    
    def __init__(self, rada, fraction_class, deputat_class, db_class):
        super(UkraineConsole, self).__init__(rada, fraction_class, deputat_class, db_class)
        self.help_text = """
                Введіть 1, щоб додати фракцію
                Введіть 2, щоб видалити фракцію
                Введіть 3, щоб очистити фракцію
                Введіть 4, щоб вивести фракції
                Введіть 5, щоб видалити фракцію з ВР
                Введіть 6, щоб додати депутата у фракцію
                Введіть 7, щоб видалити депутата із фракції
                Введіть 8, щоб вивести список хабарників у фракції
                Введіть 9, щоб вивести список  хабарників у раді
                Введіть 10, щоб вивести найбільшого хабарника у раді
                Введіть 11, щоб вивести найбільшого хабарника у фракції.
                Введіть 12, щоб перевірити чи є дупутат у фракції
                Введіть 13, щоб перевірити чи є депутат у раді
                Введіть 14, дати хабара депутату!
                Введіть 15, вивести всіх депутатів у ВР
                Введіть 16, записати у БД
                Введіть 17, зчитати з БД

                Введіть 0, щоб вийти із програми.
                : """

    def run(self):

        while True:

            choice = input(self.help_text)
            try:
                choice = int(choice)
            except ValueError:
                print('Enter a number from the list')
                continue

            if choice == 1:
                fr_name = input('Please enter a fraction name: ')
                self.rada.add_fraction(self.fraction_class(fr_name, self.deputat_class))
            elif choice == 2:
                fr_name = input('Please enter a fraction name to be deleted: ')
                self.rada.remove_fraction(self.fraction_class(fr_name, self.deputat_class))
            elif choice == 3:
                fr_name = input('Please enter fraction name to work with: ')
                self.fraction_class(fr_name, self.deputat_class).clean_out_deputies()
            elif choice == 4:
                self.rada.print_all_fractions()
            elif choice == 5:
                fr_name = input('Please enter a fraction name to be deleted: ')
                self.rada.remove_fraction(self.fraction_class(fr_name, self.deputat_class))
            elif choice == 6:
                fr_name = input('Please enter fraction name to work with: ')
                fr = self.fraction_class(fr_name, self.deputat_class)
                if fr not in self.rada.fractions:
                    self.rada.fractions.add(fr)
                for fraction in self.rada.fractions:
                    if fraction == fr:
                        fraction.add_deputy()
            elif choice == 7:
                fr_name = input('Please enter fraction name to work with: ')
                fr = self.fraction_class(fr_name, self.deputat_class)
                dp = fr.add_deputy()

                self.rada.remove_deputy_from_fraction(fr, dp)
            elif choice == 8:
                fr_name = input('Please enter fraction name to work with: ')
                fr = self.fraction_class(fr_name, self.deputat_class)

                if fr in self.rada.fractions:
                    for fraction in self.rada.fractions:
                        if fraction == fr:
                            fraction.find_all_bribe_takers()
            elif choice == 9:
                self.rada.print_all_bribe_takers()
            elif choice == 10:
                self.rada.higher_bribe_taker()
            elif choice == 11:
                fr_name = input('Please enter fraction name to work with: ')
                fr = self.fraction_class(fr_name, self.deputat_class)

                if fr in self.rada.fractions:
                    for fraction in self.rada.fractions:
                        if fraction == fr:
                            fraction.find_highest_bribe_taker()

            elif choice == 12:
                fr_name = input('Please enter fraction name to work with: ')
                fr = self.fraction_class(fr_name, self.deputat_class)
                dp = self.deputat_class.describe_deputy()
                if dp:
                    fr.is_participant(self.rada, dp)
            elif choice == 13:
                dp = self.deputat_class.describe_deputy()
                if dp:
                    self.rada.is_present_in_self.rada(dp)
            elif choice == 14:
                dp = self.deputat_class.describe_deputy()
                if dp:
                    dp.give_tribute()
                    self.rada.update_deputy(dp)
            elif choice == 15:
                self.rada.print_all_deputies()
            elif choice == 16:
                self.db.save_to_db(self.rada)
            elif choice == 17:
                self.rada = self.db.load_from_db()
            elif choice == 0:
                break
    
    
class PolandConsole(Console):

    def __init__(self, rada, fraction_class, deputat_class, db_class):
        super(PolandConsole, self).__init__(rada, fraction_class, deputat_class, db_class)
        self.help_text = """
                Введіть 1, щоб додати фракцію
                Введіть 2, щоб видалити фракцію
                Введіть 3, щоб очистити фракцію
                Введіть 4, щоб вивести фракції
                Введіть 5, щоб видалити фракцію з ВР
                Введіть 6, щоб додати депутата у фракцію
                Введіть 7, щоб видалити депутата із фракції
                Введіть 12, щоб перевірити чи є дупутат у фракції
                Введіть 13, щоб перевірити чи є депутат у раді
                Введіть 15, вивести всіх депутатів у ВР
                Введіть 16, записати у БД
                Введіть 17, зчитати з БД

                Введіть 0, щоб вийти із програми.
                : """

    def run(self):

        while True:

            choice = input(self.help_text)
            try:
                choice = int(choice)
            except ValueError:
                print('Enter a number from the list')
                continue

            if choice == 1:
                fr_name = input('Please enter a fraction name: ')
                self.rada.add_fraction(self.fraction_class(fr_name, self.deputat_class))
            elif choice == 2:
                fr_name = input('Please enter a fraction name to be deleted: ')
                self.rada.remove_fraction(self.fraction_class(fr_name, self.deputat_class))
            elif choice == 3:
                fr_name = input('Please enter fraction name to work with: ')
                self.fraction_class(fr_name, self.deputat_class).clean_out_deputies()
            elif choice == 4:
                self.rada.print_all_fractions()
            elif choice == 5:
                fr_name = input('Please enter a fraction name to be deleted: ')
                self.rada.remove_fraction(self.fraction_class(fr_name, self.deputat_class))
            elif choice == 6:
                fr_name = input('Please enter fraction name to work with: ')
                fr = self.fraction_class(fr_name, self.deputat_class)
                if fr not in self.rada.fractions:
                    self.rada.fractions.add(fr)
                for fraction in self.rada.fractions:
                    if fraction == fr:
                        fraction.add_deputy()
            elif choice == 7:
                fr_name = input('Please enter fraction name to work with: ')
                fr = self.fraction_class(fr_name, self.deputat_class)
                dp = fr.add_deputy()

                self.rada.remove_deputy_from_fraction(fr, dp)
            elif choice == 12:
                fr_name = input('Please enter fraction name to work with: ')
                fr = self.fraction_class(fr_name, self.deputat_class)
                dp = self.deputat_class.describe_deputy()
                if dp:
                    fr.is_participant(self.rada, dp)
            elif choice == 13:
                dp = self.deputat_class.describe_deputy()
                if dp:
                    self.rada.is_present_in_self.rada(dp)
            elif choice == 15:
                self.rada.print_all_deputies()
            elif choice == 16:
                self.db.save_to_db(self.rada)
            elif choice == 17:
                self.rada = self.db.load_from_db()
            elif choice == 0:
                break
